extern crate ctrlc;
use shakmaty::fen::Fen;
use shakmaty::{Chess, Color, Outcome, Position, Setup};
use std::io;
use std::io::{BufRead, BufReader};
use std::os::unix::net::UnixStream;
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::{Arc, Mutex};
use std::thread;
use std::time;

fn main() {
    let running = setupctrlc();
    let username = std::env::args().nth(1).expect("no name given");
    let public_key = std::env::args().nth(2).expect("no public key given");
    println!("Name: {}, Public key: {}", username, public_key);
    let mut stream = match UnixStream::connect("/tmp/clichess.socket") {
        Ok(sock) => sock,
        Err(_) => {
            println!("clichess daemon is not running.");
            return;
        }
    };
    //send username and public key to server
    clichess::write_to_stream(&mut stream, username);
    clichess::write_to_stream(&mut stream, public_key);
    let (client, chess) = get_connection_info_from_stream(&stream);
    //First prompt when connecting to the server
    println!(
        "Hello, {} !\n\r You're playing with the {} pieces",
        client.username,
        client.role.to_string()
    );
    //then we get the initial role of the connected client.
    let mut current_position = chess;
    let input_buffer = setup_input_buffer();
    loop {
        println!(
            "{}",
            clichess::board_representation(&current_position, clichess::to_color(client.side))
        );
        //check if game is over.
        if current_position.is_game_over() {
            break;
        }
        if clichess::is_player_turn(&client, current_position.turn()) {
            //it's the user turn, taking user input
            let input = read_user_input(running.clone(), input_buffer.clone());
            println!("trying to play {}", input);
            clichess::write_to_stream(&mut stream, String::from(input.trim()));
            if input.trim() == "exit" {
                break;
            }
        }
        //update position after playing.
        current_position = get_current_position(&stream);
    }
    match current_position.outcome() {
        None => println!("Bye"),
        Some(Outcome::Draw) => println!("Draw game."),
        Some(Outcome::Decisive { winner }) => {
            if winner == Color::White {
                println!("White has won the game.")
            } else {
                println!("Black has won the game.")
            }
        }
    }
}

fn setup_input_buffer() -> Arc<Mutex<String>> {
    let buf = Arc::new(Mutex::new(String::new()));
    let buf2 = buf.clone();

    thread::spawn(move || {
        loop {
            let mut buffer = String::new();
            //wait for user input
            io::stdin().read_line(&mut buffer).unwrap();
            {
                let mut user_input = buf2.lock().unwrap();
                if user_input.is_empty() {
                    *user_input = buffer;
                } else {
                    println!("It's not your turn !");
                }
            }
        }
    });
    buf
}

fn setupctrlc() -> Arc<AtomicBool> {
    let running = Arc::new(AtomicBool::new(true));
    let r = running.clone();
    ctrlc::set_handler(move || {
        r.store(false, Ordering::SeqCst);
    })
    .expect("Error setting Ctrl-C handler");
    running
}

fn read_user_input(running: Arc<AtomicBool>, input_buffer: Arc<Mutex<String>>) -> String {
    let mut input = String::new();
    while running.load(Ordering::SeqCst) {
        thread::sleep(time::Duration::from_millis(10));
        {
            let user_input = input_buffer.lock().unwrap();
            if user_input.is_empty() {
                continue;
            } else {
                input = user_input.clone();
                break;
            }
        }
    }
    if running.load(Ordering::SeqCst) {
        input
    } else {
        String::from("exit")
    }
}

//wait for next position from server, then return the current board.
fn get_current_position(stream: &UnixStream) -> Chess {
    let response = clichess::read_line_from_stream(&stream).expect("Server disconnected.");
    parse_position(&response)
}

fn parse_position(string: &str) -> Chess {
    let setup: Fen = string.trim().parse().expect("Invalid message from server.");
    let position: Chess = setup.position().expect("Invalid message from server.");
    position
}

fn get_connection_info_from_stream(stream: &UnixStream) -> (clichess::Client, Chess) {
    println!("Read lines from stream...");
    let mut buf = String::new();
    let mut reader = BufReader::new(stream);
    //first, client as json.
    reader
        .read_line(&mut buf)
        .expect("Server closed connection.");
    let client = serde_json::from_str(&(buf.trim())).unwrap();
    buf.clear();
    reader
        .read_line(&mut buf)
        .expect("Server closed connection.");
    let chess = parse_position(buf.trim());
    (client, chess)
}
